# -*- coding: utf-8 -*-
"""
Created on Wed May 24 15:49:58 2023

@author: heyunan
"""

import torch
from torchvision import datasets
from torchvision.transforms import ToTensor, Compose, Normalize
from torch.utils.data import DataLoader
import torch.nn as nn
import torch.nn.functional as F
import os
import torch.optim as optim
import math
import matplotlib.pyplot as plt

os.environ['KMP_DUPLICATE_LIB_OK']='TRUE'


# ******************************************
# 初始化方法
# ******************************************

init_methods = ["kaiming_uniform"]


def kaiming_normal(m):
    if type(m) == nn.Linear:
        nn.init.kaiming_normal_(m.weight, a=math.sqrt(5))
        fan_in, _ = nn.init._calculate_fan_in_and_fan_out(m.weight)
        bound = 1 / math.sqrt(fan_in)
        nn.init.normal_(m.bias, mean=0, std=bound)

    elif type(m) == nn.Conv2d:
        nn.init.kaiming_normal_(m.weight, a=math.sqrt(5))
        if m.bias is not None:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(m.weight)
            bound = 1 / math.sqrt(fan_in)
            nn.init.normal_(m.bias, mean=0, std=bound)


def kaiming_uniform(m):
    if type(m) == nn.Linear:
        nn.init.kaiming_uniform_(m.weight, a=math.sqrt(5))
        fan_in, _ = nn.init._calculate_fan_in_and_fan_out(m.weight)
        bound = 1 / math.sqrt(fan_in)
        nn.init.uniform_(m.bias, -bound, bound)

    elif type(m) == nn.Conv2d:
        nn.init.kaiming_uniform_(m.weight, a=math.sqrt(5))
        if m.bias is not None:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(m.weight)
            bound = 1 / math.sqrt(fan_in)
            nn.init.uniform_(m.bias, -bound, bound)


def default_normal(m, mean=0.0, std=0.01):
    if isinstance(m, nn.Linear) or isinstance(m, nn.Conv2d):
        nn.init.normal_(m.weight, mean=mean, std=std)
        if m.bias is not None:
            nn.init.normal_(m.bias, mean=mean, std=std)


def default_uniform(m, a=0.0, b=0.1):
    if isinstance(m, nn.Linear) or isinstance(m, nn.Conv2d):
        nn.init.uniform_(m.weight, a=a, b=b)
        if m.bias is not None:
            nn.init.uniform_(m.bias, a=a, b=b)


# ******************************************
# 定义神经网络架构
# ******************************************

class Net(nn.Module):
    def __init__(self, init_method):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 32, 3, 1)
        self.conv2 = nn.Conv2d(32, 64, 3, 1)
        self.dropout1 = nn.Dropout(0.25)
        self.dropout2 = nn.Dropout(0.5)
        self.fc1 = nn.Linear(9216, 128)
        self.fc2 = nn.Linear(128, 10)

        # 显式初始化权重和偏置

        if init_method == 'kaiming_uniform':
            self.apply(kaiming_uniform)
            print('Using initialization method: kaiming_uniform')

        if init_method == 'kaiming_normal':
            self.apply(kaiming_normal)
            print('Using initialization method: kaiming_normal')

        if init_method == 'default_normal':
            self.apply(default_normal)
            print('Using initialization method: default_normal')

        if init_method == 'default_uniform':
            self.apply(default_uniform)
            print('Using initialization method: default_uniform')

    def forward(self, x):
        x = self.conv1(x)
        x = F.relu(x)
        x = self.conv2(x)
        x = F.relu(x)
        x = F.max_pool2d(x, 2)
        x = self.dropout1(x)
        x = torch.flatten(x, 1)
        x = self.fc1(x)
        x = F.relu(x)
        x = self.dropout2(x)
        x = self.fc2(x)
        output = F.log_softmax(x, dim=1)
        return output


def train(model, device, train_loader, optimizer, init_method, lossess, log_interval=100):
    model.train()

    for batch_idx, (data, target) in enumerate(train_loader):

        if ((epoch - 1) * len(train_loader) + batch_idx) % log_interval == 0:

            # Create a new folder to store the weights for this init method
            folder_name = 'weights/{}'.format(init_method)

            if not os.path.exists(folder_name):
                os.makedirs(folder_name, exist_ok=True)

            torch.save(network.state_dict(), os.path.join(folder_name, 'iterations_{}_params.pt'.format(
                (epoch - 1) * len(train_loader) + batch_idx)))

        data, target = data.to(device), target.to(device)
        optimizer.zero_grad()
        output = model(data)
        loss = F.nll_loss(output, target)

        losses.append(loss.item())  # 添加损失值到列表
        loss.backward()
        optimizer.step()

        if ((epoch - 1) * len(train_loader) + batch_idx) % log_interval == 0:
            print('Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                epoch, batch_idx * len(data), len(train_loader.dataset),
                       100. * batch_idx / len(train_loader), loss.item()))


def test(model, device, test_loader):
    model.eval()
    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to(device), target.to(device)
            output = model(data)
            test_loss += F.nll_loss(output, target, reduction='sum').item()  # sum up batch loss
            pred = output.argmax(dim=1, keepdim=True)  # get the index of the max log-probability
            correct += pred.eq(target.view_as(pred)).sum().item()

    test_loss /= len(test_loader.dataset)

    print('Average loss: {:.4f}, Accuracy: {}/{} ({:.2f}%) in Epoch {}\n'.format(
        test_loss, correct, len(test_loader.dataset),
        100. * correct / len(test_loader.dataset), epoch)
    )


def reset_all_weights(model):
    for layer in model.children():
        if hasattr(layer, 'reset_parameters'):
            layer.reset_parameters()


training_data = datasets.MNIST(
    root="mnist_data",
    train=True,
    download=True,
    transform=Compose([ToTensor(),
                       Normalize((0.1307,), (0.3081,))]))

test_data = datasets.MNIST(
    root="mnist_data",
    train=False,
    download=True,
    transform=Compose([ToTensor(),
                       Normalize((0.1307,), (0.3081,))]))

train_loader = DataLoader(training_data, batch_size=64, shuffle=True)
test_loader = DataLoader(test_data, batch_size=64, shuffle=True)

epochs = 20  # number of epochs you want to train
# models = 1 # number of models you want to save
learning_rate = 0.01

# setting device on GPU if available, else CPU
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
print('Using device:', device)

runs_per_method = 1

for init_method in init_methods:

    for run in range(runs_per_method):
        #  for run in [0]:
        run_folder = f'{init_method}_run{run + 1}'  # 添加run编号到文件夹名称
        network = Net(init_method).to(device)
        optimizer = optim.SGD(network.parameters(), lr=learning_rate, momentum=0.5)
        losses = []

        for epoch in range(1, epochs + 1):
            # 将run_folder传入train函数，以便保存权重到正确的文件夹
            train(network, device, train_loader, optimizer, run_folder, losses)
            print(f"Validation on epoch {epoch}:")
            test(network, device, test_loader)

        # Create a new folder to store the weights for this init method
        folder_name = f'weights/{run_folder}'

        plt.plot(losses)
        plt.xlabel('Iteration')
        plt.ylabel('Loss')
        plt.title('Training Loss')
        plt.savefig(os.path.join(folder_name, 'loss_curve.pdf'))  # 保存损失曲线为图片
        plt.clf()

        reset_all_weights(network)

