import os
import torch
import torch.nn as nn
from tqdm import tqdm
import matplotlib.pyplot as plt
import torch.nn.functional as F
from torchvision import datasets
from torchvision.transforms import Compose, ToTensor, Normalize
from torch.utils.data import DataLoader
import pickle

# Define the neural network structure
class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        # First convolution layer
        self.conv1 = nn.Conv2d(1, 32, 3, 1)
        # Second convolution layer
        self.conv2 = nn.Conv2d(32, 64, 3, 1)
        # Dropout layers to prevent overfitting
        self.dropout1 = nn.Dropout(0.25)
        self.dropout2 = nn.Dropout(0.5)
        # Fully connected layers
        self.fc1 = nn.Linear(9216, 128)
        self.fc2 = nn.Linear(128, 10)

    def forward(self, x):
        # Forward pass through the network
        x = F.relu(self.conv1(x))
        x = F.relu(self.conv2(x))
        x = F.max_pool2d(x, 2)
        x = self.dropout1(x)
        x = torch.flatten(x, 1)
        x = F.relu(self.fc1(x))
        x = self.dropout2(x)
        x = self.fc2(x)
        output = F.log_softmax(x, dim=1)
        return output

# Load training and test data
training_data = datasets.MNIST(
    root="mnist_data",
    train=True,
    download=True,
    transform=Compose([ToTensor(), Normalize((0.1307,), (0.3081,))])
)

test_data = datasets.MNIST(
    root="mnist_data",
    train=False,
    download=True,
    transform=Compose([ToTensor(), Normalize((0.1307,), (0.3081,))])
)

# DataLoader for training and test data
train_loader = DataLoader(training_data, batch_size=64, shuffle=True)
test_loader = DataLoader(test_data, batch_size=64, shuffle=True)

# Check for device (GPU/CPU)
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
print(device)

# Function to compute loss
def compute_loss(net, loader, device):
    net.eval()  # Set the network to evaluation mode
    total_loss = 0
    with torch.no_grad():  # No need to calculate gradients
        for data, target in loader:
            data, target = data.to(device), target.to(device)
            output = net(data)
            # Use Negative Log Likelihood Loss
            total_loss += F.nll_loss(output, target, reduction='sum').item()
    average_loss = total_loss / len(loader.dataset)
    return average_loss

# Function to plot the loss curve
def plot_loss_curve(iterations, losses, dir_name):
    plt.figure(figsize=(10, 5))
    plt.title(f"Loss curve for {dir_name}")
    plt.plot(iterations, losses, label='Loss')
    plt.xlabel('Iterations')
    plt.ylabel('Loss')
    plt.legend()
    plt.savefig(os.path.join(weights_dir, dir_name, 'loss_curve.png'))  # Save the plot
    plt.close()

# Directory for weights
weights_dir = "./weights"
dir_names = [name for name in os.listdir(weights_dir) if os.path.isdir(os.path.join(weights_dir, name))]

# Initialize the network
net = Net().to(device)

# Process each directory
for d in dir_names:
    d_path = os.path.join(weights_dir, d)
    pickle_path = os.path.join(d_path, 'loss_data.pickle')

    iterations, losses = [], []
    files = os.listdir(d_path)

    # Load weights and compute loss for each saved model
    for file_name in tqdm(files, desc=f"Processing {d}"):
        if "_params.pt" in file_name:
            iteration = int(file_name.split('_')[1])
            iterations.append(iteration)

            net.load_state_dict(torch.load(os.path.join(d_path, file_name)))
            loss = compute_loss(net, test_loader, device)

            losses.append(loss)

    iter_loss_pairs = list(zip(iterations, losses))
    iter_loss_pairs.sort()
    iterations, losses = zip(*iter_loss_pairs)
    plot_loss_curve(iterations, losses, d)

    # Save the iteration-loss pairs
    with open(pickle_path, 'wb') as f:
        pickle.dump(iter_loss_pairs, f)
