import os
import torch
import torch.nn.functional as F
import numpy as np
from sklearn.metrics import roc_curve, auc
import matplotlib.pyplot as plt
from torchvision import datasets, transforms
from torch.utils.data import DataLoader
import torch.nn as nn
from tqdm import tqdm
from sklearn.preprocessing import label_binarize
import re

class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        # First convolution layer
        self.conv1 = nn.Conv2d(1, 32, 3, 1)
        # Second convolution layer
        self.conv2 = nn.Conv2d(32, 64, 3, 1)
        # Dropout layers to prevent overfitting
        self.dropout1 = nn.Dropout(0.25)
        self.dropout2 = nn.Dropout(0.5)
        # Fully connected layers
        self.fc1 = nn.Linear(9216, 128)
        self.fc2 = nn.Linear(128, 10)

    def forward(self, x):
        # Forward pass through the network
        x = F.relu(self.conv1(x))
        x = F.relu(self.conv2(x))
        x = F.max_pool2d(x, 2)
        x = self.dropout1(x)
        x = torch.flatten(x, 1)
        x = F.relu(self.fc1(x))
        x = self.dropout2(x)
        x = self.fc2(x)
        output = F.log_softmax(x, dim=1)
        return output


# Load the test data
transform = transforms.Compose([
    transforms.ToTensor(),
    transforms.Normalize((0.1307,), (0.3081,))
])
test_dataset = datasets.MNIST(root='./mnist_data', train=False, download=True, transform=transform)
test_loader = DataLoader(test_dataset, batch_size=64, shuffle=False)

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
net = Net().to(device)

weights_dir = "./weights"
sub_dirs = [d for d in os.listdir(weights_dir) if os.path.isdir(os.path.join(weights_dir, d))]

# 创建一个大图和两个子图
fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(20, 8))
fig.suptitle('Macro-Average ROC Curve and Zoomed In View', fontsize=20)

for sub_dir in sub_dirs:
    dir_path = os.path.join(weights_dir, sub_dir)
    weight_files = os.listdir(dir_path)

    # 使用正则表达式提取迭代次数，并进行排序
    weight_files_with_iterations = []
    for weight_file in weight_files:
        match = re.match(r'iterations_(\d+)_params.pt', weight_file)
        if match:
            iteration_number = int(match.group(1))  # 将迭代次数转换为整数
            weight_files_with_iterations.append((iteration_number, weight_file))

    # 根据迭代次数对文件进行排序
    weight_files_sorted = sorted(weight_files_with_iterations, key=lambda x: x[0])

    # 现在weight_files_sorted已经按迭代次数排序了，我们可以继续之前的逻辑
    for iteration_number, weight_file in tqdm(weight_files_sorted, desc=f'Loading weights from {sub_dir}', leave=False):
        weight_path = os.path.join(dir_path, weight_file)
        net.load_state_dict(torch.load(weight_path, map_location=device))
        net.eval()

        all_preds = np.array([])
        all_targets = np.array([])

        with torch.no_grad():
            for data, target in test_loader:
                data = data.to(device)
                output = net(data)
                probs = torch.softmax(output, dim=1)
                all_preds = np.append(all_preds, probs.cpu().numpy(),
                                          axis=0) if all_preds.size else probs.cpu().numpy()
                all_targets = np.append(all_targets, target.cpu().numpy(), axis=0)

        all_targets_bin = label_binarize(all_targets, classes=range(10))

        fpr = dict()
        tpr = dict()
        roc_auc = dict()
        for i in range(10):
             fpr[i], tpr[i], _ = roc_curve(all_targets_bin[:, i], all_preds[:, i])
             roc_auc[i] = auc(fpr[i], tpr[i])

        all_fpr = np.unique(np.concatenate([fpr[i] for i in range(10)]))
        mean_tpr = np.zeros_like(all_fpr)
        for i in range(10):
            mean_tpr += np.interp(all_fpr, fpr[i], tpr[i])
        mean_tpr /= 10
        roc_auc["macro"] = auc(all_fpr, mean_tpr)

        # 绘制原始ROC曲线
        ax1.plot(all_fpr, mean_tpr, lw=2, label=f'Iteration {iteration_number} ')
        # 绘制放大的ROC曲线
        ax2.plot(all_fpr, mean_tpr, lw=2, label=f'Iteration {iteration_number} ')

ax1.plot([0, 1], [0, 1], color='navy', lw=2, linestyle='--')
ax1.set_xlim([0.0, 1.0])
ax1.set_ylim([0.0, 1.05])
ax1.set_xlabel('False Positive Rate', fontsize=14)
ax1.set_ylabel('True Positive Rate', fontsize=14)
ax1.legend(loc="lower right", fontsize='medium')
ax1.set_title('ROC Curve', fontsize=16)


ax2.set_xlim(0, 0.2)
ax2.set_ylim(0.8, 1.0)
ax2.plot([0, 1], [0, 1], color='navy', lw=2, linestyle='--')
ax2.set_xlabel('False Positive Rate', fontsize=14)
ax2.set_ylabel('True Positive Rate', fontsize=14)
ax2.legend(loc="lower right", fontsize='medium')
ax2.set_title('Zoomed ROC Curve', fontsize=16)

plt.savefig(os.path.join(weights_dir, sub_dir, 'macro_avg_roc_curve.png'))  # Save the plot
plt.close()
